var BernsteinCoefficients = new Map();
BernsteinCoefficients.set(3,[[1,3,3,1],[3,6,3],[3,3],[1]]);
BernsteinCoefficients.set(4,[[1,4,6,4,1],[4,12,12,4],[6,12,6],[4,4],[1]]);
var pascal_triangle = new Array();

function TriangulationToArray(edges,triangulation,cubic_cp,quartic_cp)
{
	var triangles = [];
	for(var i=0;i<triangulation.length;i+=3)
	{
		triangles.push({"edges": [], "q" : [new Array(4),new Array(4),new Array(4)], "z": new Array(7) , "x": new Array(3)});
		const last_index = triangles.length - 1;
		for(var j=0;j<3;j++)
		{
			var dir = 1;
			var edge_first = triangulation[i+j];
			var edge_second = triangulation[i+((j+1)%3)];
			if(edge_first > edge_second)
			{
				dir = -1;
				var tmp = edge_first;  edge_first = edge_second; edge_second = tmp;
			}
			// q[][0]
			triangles[last_index].edges[j] = {"index" : edges.findIndex(index=>index.first == edge_first && index.second == edge_second), "dir" : dir};
			
			triangles[last_index].q[j][0] =  dir > 0 ? cubic_cp[triangles[last_index].edges[j].index][0] : cubic_cp[triangles[last_index].edges[j].index][3];
		}
	}
	// q[][1]
	for(var i=0;i<triangles.length;i++)
	{
		for(var j=0;j<3;j++)
		{
			const u1 = quartic_cp[triangles[i].edges[j].index][triangles[i].edges[j].dir > 0 ? 3 : 1];
			const u2 = quartic_cp[triangles[i].edges[(j+1)%3].index][triangles[i].edges[(j+1)%3].dir < 0 ? 3 : 1];
			const u0 = triangles[i].q[(j+1)%3][0];
			triangles[i].q[(j+1)%3][1] = num_vec_mul(1/3,vec_add(vec_add(u0,u1),u2));
		}
	}
	return triangles;
}

function FindAlphaBeta(edges,triangles,quartic_cp)
{  
	for(let i=0;i<edges.length;i++)
	{
		var neighbors = [];
		for(let j=0;j<triangles.length;j++)
		{
			const pos = triangles[j].edges.findIndex(x=>x.index == i);
			if(pos!= -1)
			{
				neighbors.push({"tri_index" : j , "edge_index" : pos});
				if(neighbors.length ==2)
					break;
			}
		}
		if(neighbors.length == 2)
		{
			if(triangles[neighbors[0].tri_index].edges[neighbors[0].edge_index].dir < 0)
			{
				neighbors = neighbors.reverse();
			}
			var q10 = quartic_cp[triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+2)%3].index][triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+2)%3].dir > 0 ? 3 : 1];
			var nq10 = quartic_cp[triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+1)%3].index][triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+1)%3].dir < 0 ? 3 : 1];
			var q20 = quartic_cp[triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+1)%3].index][triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+1)%3].dir < 0 ? 3 : 1];
			var nq20 = quartic_cp[triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+2)%3].index][triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+2)%3].dir > 0 ? 3 : 1];
			
			var tmp = DiagonalCoef(nq10,quartic_cp[i][1],q10,quartic_cp[i][0]);
			edges[i].alpha_left = tmp.alpha;
			edges[i].beta_left = tmp.beta;
			var tmp1 = DiagonalCoef(nq20,quartic_cp[i][4],q20,quartic_cp[i][3]);
			edges[i].alpha_right = tmp1.alpha;
			edges[i].beta_right = tmp1.beta;
			
	/*		console.log(vec_add(num_vec_mul(edges[i].alpha_right,q20),num_vec_mul(1-edges[i].alpha_right,nq20)),
						vec_add(num_vec_mul(edges[i].beta_right,quartic_cp[i][3]),num_vec_mul(1-edges[i].beta_right,quartic_cp[i][4])));
		*/	
		}
	}

}

function UpdateAlphaBeta(edges,triangles,quartic_cp)
{  
	for(let i=0;i<edges.length;i++)
	{
		var neighbors = [];
		for(let j=0;j<triangles.length;j++)
		{
			const pos = triangles[j].edges.findIndex(x=>x.index == i);
			if(pos!= -1)
			{
				neighbors.push({"tri_index" : j , "edge_index" : pos});
				if(neighbors.length ==2)
					break;
			}
		}
		if(neighbors.length == 2 && (triangles[neighbors[0].tri_index].splitting === true || triangles[neighbors[1].tri_index].splitting === true))
		{
			if(triangles[neighbors[0].tri_index].edges[neighbors[0].edge_index].dir < 0)
			{
				neighbors = neighbors.reverse();
			}
			var q10 = quartic_cp[triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+2)%3].index][triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+2)%3].dir > 0 ? 3 : 1];
			var nq10 = quartic_cp[triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+1)%3].index][triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+1)%3].dir < 0 ? 3 : 1];
			var q20 = quartic_cp[triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+1)%3].index][triangles[neighbors[0].tri_index].edges[(neighbors[0].edge_index+1)%3].dir < 0 ? 3 : 1];
			var nq20 = quartic_cp[triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+2)%3].index][triangles[neighbors[1].tri_index].edges[(neighbors[1].edge_index+2)%3].dir > 0 ? 3 : 1];
			if(triangles[neighbors[0].tri_index].splitting === true)
			{
				q10 = num_vec_mul(1/3,vec_add(q10,vec_add(quartic_cp[i][0],quartic_cp[i][1])));
				q20 = num_vec_mul(1/3,vec_add(q20,vec_add(quartic_cp[i][3],quartic_cp[i][4])));
			}
			if(triangles[neighbors[1].tri_index].splitting === true)
			{
				nq10 = num_vec_mul(1/3,vec_add(nq10,vec_add(quartic_cp[i][1],quartic_cp[i][0])));
				nq20 = num_vec_mul(1/3,vec_add(nq20,vec_add(quartic_cp[i][3],quartic_cp[i][4])));
			}
			var tmp = DiagonalCoef(nq10,quartic_cp[i][1],q10,quartic_cp[i][0]);			
			edges[i].alpha_left = tmp.alpha;
			edges[i].beta_left = tmp.beta;
			var tmp1 = DiagonalCoef(nq20,quartic_cp[i][4],q20,quartic_cp[i][3]);
			edges[i].alpha_right = tmp1.alpha;
			edges[i].beta_right = tmp1.beta;
		
		}
	}

}

function FindTS(points,edges,triangles,triangulation,quartic_cp)
{
	var marked = (new Array(points.length)).fill(0);
	var cond5 = (new Array(points.length)).fill(false);
	var TS = new Set();
	for(let i=0;i<points.length;i++)
	{
		var incident_e = Array.from(edges.keys()).filter(e=>edges[e].first == i || edges[e].second == i);
		points[i].interior = false;
		points[i].degree = incident_e.length;
		if(incident_e.length > 0 && incident_e.every(e=>edges[e].alpha_left !== undefined && edges[e].alpha_left !== null && !isNaN(edges[e].alpha_left)))
		{ // interior vertex
			points[i].interior = true;
			var edges_info = [];
			var last_edge = {'index': incident_e[0], 'dir': 1};
			if(edges[incident_e[0]].first == i)
				edges_info.push({'index': incident_e[0],'q': quartic_cp[incident_e[0]], 'a0':edges[incident_e[0]].alpha_left, 'a1': edges[incident_e[0]].alpha_right, 'b0':edges[incident_e[0]].beta_left, 'b1':edges[incident_e[0]].beta_right});
			else
			{
				edges_info.push({'index': incident_e[0], 'q':quartic_cp[incident_e[0]].slice().reverse(), 'a0':1-edges[incident_e[0]].alpha_right, 'a1':1-edges[incident_e[0]].alpha_left, 'b0':1-edges[incident_e[0]].beta_right, 'b1':1-edges[incident_e[0]].beta_left});
				last_edge.dir = -1;
			}			
			while(edges_info.length != incident_e.length)
			{
				for(let j=0;j<triangles.length;j++)
				{
					const pos = triangles[j].edges.findIndex(e=>e.index == last_edge.index && e.dir == last_edge.dir);
					if(pos!= -1)
					{
						last_edge.index = triangles[j].edges[(pos+2)%3].index;
						last_edge.dir = 1;
						if(edges[last_edge.index].first == i)
							edges_info.push({'index': last_edge.index ,'q': quartic_cp[last_edge.index], 'a0':edges[last_edge.index].alpha_left, 'a1': edges[last_edge.index].alpha_right, 'b0':edges[last_edge.index].beta_left, 'b1': edges[last_edge.index].beta_right});
						else
						{
							edges_info.push({'index': last_edge.index,'q': quartic_cp[last_edge.index].slice().reverse(), 'a0':1-edges[last_edge.index].alpha_right, 'a1':1-edges[last_edge.index].alpha_left, 'b0':1-edges[last_edge.index].beta_right, 'b1':1-edges[last_edge.index].beta_left});
							last_edge.dir = -1;
						}		
						break;
					}					
				}	
			}
			var mark_e = edges_info.filter(e=>!isZero(e.a0-e.a1)).map(e=>e.index);			
			mark_e.forEach(function(index){ 
				marked[edges[index].first] = 1; 
				marked[edges[index].second] = 1;
			});
			Array.from(triangles.keys()).filter(index=>triangles[index].edges.some(e_i=>mark_e.includes(e_i.index))).forEach(function(index){TS.add(index);});		
			cond5[i] = true;
			if(points[i].degree%2 == 0)
			{		
				var s = new Array(edges_info.length);
				for(let k=0;k<edges_info.length;k++)
				{
					s[k] = 
					vec_sub(
					vec_add(num_vec_mul(edges_info[(k+1)%s.length].b0,edges_info[(k+1)%s.length].q[1]), num_vec_mul(1-edges_info[(k+1)%s.length].b0,edges_info[(k+1)%s.length].q[2])),
					num_vec_mul(1/3,
					vec_sub(
					vec_add(num_vec_mul(edges_info[(k+1)%s.length].a1,edges_info[(k+2)%s.length].q[1]), num_vec_mul(1-edges_info[(k+1)%s.length].a1,edges_info[k].q[1])),
					vec_add(num_vec_mul(edges_info[(k+1)%s.length].b1,edges_info[k].q[0]), num_vec_mul(1-edges_info[(k+1)%s.length].b1,edges_info[(k+1)%s.length].q[1]))
					)));
				}
				var sum = {'x': 0, 'y': 0, 'z':0};
				for(let m=0;m<edges_info.length-1;m++)
				{
					var coef = m % 2 == 0 ? -1 : 1;
					for(let k=0;k<edges_info.length;k++)
					{
						if(k < m+1)
							coef = coef * edges_info[k].a0;
						else if(k > m+1)
							coef = coef * (1-edges_info[k].a0);					
					}
					sum = vec_add(sum,num_vec_mul(coef,s[m]));
				}
				var coef = 1;
				for(let k=1;k<edges_info.length;k++)
					coef = coef * (1-edges_info[k].a0);	
				if(!isZero(len(vec_add(sum,num_vec_mul(coef,s[edges_info.length-1])))))						
					cond5[i] = false;

			}
		}
	}
	if(TS.size != triangles.length)
	{
		var V2 = new Set(Array.from(points.keys()).filter(i=>marked[i] == 0 && points[i].degree%2==0 && cond5[i] == false));	
	//	console.log(Array.from(V2.values()),Array.from(points.keys()).filter(x=>points[x].interior == true));
		var Q = Array.from(triangles.keys()).map(index=>triangles[index].edges.reduce((p,e_i)=>p+(V2.has(edges[e_i.index].first) ? 1:0) + (V2.has(edges[e_i.index].second)? 1:0),0));
	//	console.log(TS);
		while(Q.some(w=>w >0))
		{
//			console.log("-----------------------------------------");
//			console.log(Q);
			var max_t = {'index' : 0, 'max_21' : null, 'min_cos': null,'S': null};
			for(let i=1;i<Q.length;i++)
			{
				if(Q[i] > Q[max_t.index])
				{
					max_t.index = i;
					max_t.max_21 = null;
					max_t.min_cos = null;
					max_t.S = null;
				}
				else if(Q[i] == Q[max_t.index])
				{
					const tri_i_pts = [triangulation[3*i],triangulation[3*i+1],triangulation[3*i+2]];					
					const tri_max_pts = [triangulation[3*max_t.index],triangulation[3*max_t.index+1],triangulation[3*max_t.index+2]];					
					var i_21 = tri_i_pts.map(v=>edges.filter(e=>(e.first == v && points[e.second].interior == false )|| (e.second == v) && points[e.first].interior == false).length).reduce((p,n)=>p+n,0);
//					console.log(i,i_21,tri_i_pts,tri_i_pts.map(v=>edges.filter(e=>(e.first == v && points[e.second].interior == false )|| (e.second == v) && points[e.first].interior == false)));
					
					if(max_t.max_21 == null)
					{
						max_t.max_21 = tri_max_pts.map(v=>edges.filter(e=>(e.first == v && points[e.second].interior == false )|| (e.second == v) && points[e.first].interior == false).length).reduce((p,n)=>p+n,0);
					}
					if(i_21 < max_t.max_21)
					{
						max_t.index = i;
						max_t.max_21 = i_21;
						max_t.min_cos = null;
						max_t.S = null;
					}
					else if(i_21 == max_t.max_21)
					{
						var i_ab =  vec_sub(points[tri_i_pts[1]],points[tri_i_pts[0]]);
						i_ab.z = 0;
						var i_bc =  vec_sub(points[tri_i_pts[2]],points[tri_i_pts[1]]);
						i_bc.z = 0;
						var i_ca =  vec_sub(points[tri_i_pts[0]],points[tri_i_pts[2]]);
						i_ca.z = 0;					
						const i_min_cos= [cos(num_vec_mul(-1,i_ca),i_ab),cos(num_vec_mul(-1,i_ab),i_bc),cos(num_vec_mul(-1,i_bc),i_ca)].reduce((p,n)=> p > n ? p : n);						
						if(max_t.min_cos == null)
						{
							var max_t_ab =  vec_sub(points[tri_max_pts[1]],points[tri_max_pts[0]]);
							max_t_ab.z = 0;
							var max_t_bc =  vec_sub(points[tri_max_pts[2]],points[tri_max_pts[1]]);
							max_t_bc.z = 0;
							var max_t_ca =  vec_sub(points[tri_max_pts[0]],points[tri_max_pts[2]]);
							max_t_ca.z = 0;							
							max_t.min_cos = [cos(num_vec_mul(-1,max_t_ca),max_t_ab),cos(num_vec_mul(-1,max_t_ab),max_t_bc),cos(num_vec_mul(-1,max_t_bc),max_t_ca)].reduce((p,n)=> p > n ? p : n);						
						}
						
						if(i_min_cos < max_t.min_cos && !isZero(i_min_cos-max_t.min_cos))
						{
							max_t.index = i;
							max_t.max_21 = i_21;
							max_t.min_cos = i_min_cos;
							max_t.S = null;
						}						
						else if(isZero(i_min_cos-max_t.min_cos))
						{
							const i_pts = tri_i_pts.map(index=>new Object({'x':points[index].x,'y':points[index].y,'z':0}));
							const i_S = TriArea(i_pts[0],i_pts[1],i_pts[2]);
							if(max_t.S == null)
							{
								const max_pts = tri_max_pts.map(index=>new Object({'x':points[index].x,'y':points[index].y,'z':0}));
								max_t.S = TriArea(max_pts[0],max_pts[1],max_pts[2]);
							}
							if(i_S < max_t.S)
							{
								max_t.index = i;
								max_t.max_21 = i_21;
								max_t.min_cos = i_min_cos;
								max_t.S = i_S;
							}
						}		
					}
				}
			}
//			console.log([triangulation[3*max_t.index],triangulation[3*max_t.index+1],triangulation[3*max_t.index+2]]);
//			console.log(max_t,[triangulation[3*max_t.index],triangulation[3*max_t.index+1],triangulation[3*max_t.index+2]].map(p=>points[p]));
			TS.add(max_t.index);
			[triangulation[3*max_t.index],triangulation[3*max_t.index+1],triangulation[3*max_t.index+2]].forEach(i=>V2.delete(i));
			Q = Array.from(triangles.keys()).map(index=>triangles[index].edges.reduce((p,e_i)=>p+(V2.has(edges[e_i.index].first) ? 1:0) + (V2.has(edges[e_i.index].second)? 1:0),0));
			//console.log(V2);
		}
		
	}
//	console.log(TS);
	for(let i =0;i<triangles.length;i++)
	{
		if(TS.has(i))
			triangles[i].splitting = true;
		else
			triangles[i].splitting = false;
	}
	return TS;
}

function Alg_1_Step_3_4(points,edges,triangles,cubic_cp,quartic_cp)
{
	for(let i=0;i<points.length;i++)
	{
		if(points[i].interior == true)
		{
			const first_split_i = triangles.findIndex(t=>t.splitting == true && t.edges.some(e=>edges[e.index].first == i || edges[e.index].second == i));
			if(first_split_i == -1)
			{ // vsi4ki ne sa za splitvane togava direktno sus sistemata
			//	console.log("za sistemata",points[i],i);			
				const incident_edges_count = edges.filter(e=>e.first == i || e.second == i).length;			
				var tris_right_edges = [{'tri_index': -1, 'edge_index': -1}];
				for(let t=0;t<triangles.length;t++)
				{
					var ready = false;
					for(let e_i = 0;e_i<3;e_i++)
					{
						if((triangles[t].edges[e_i].dir == 1 && edges[triangles[t].edges[e_i].index].first == i) || (triangles[t].edges[e_i].dir == -1 && edges[triangles[t].edges[e_i].index].second == i))
						{
							tris_right_edges[0].tri_index = t;
							tris_right_edges[0].edge_index = e_i;
							ready = true;
							break;
						}
					}
					if(ready)
						break;
				}

			//	console.log("max_size",incident_edges_count);
				while(tris_right_edges.length <= incident_edges_count)
				{					
					//console.log("da",tris_right_edges);
					var neighbor_info = GetNeighborTri(triangles,tris_right_edges[tris_right_edges.length-1].tri_index,(tris_right_edges[tris_right_edges.length-1].edge_index+2)%3);			
					if(neighbor_info.tri_index != tris_right_edges[0].tri_index)
					{ // dobavme nov triugulnik i negovoto dqsno rebro ilizato ot i
						tris_right_edges.push({'tri_index': neighbor_info.tri_index, 'edge_index': neighbor_info.edge_index});	
					}
					else
						break;
				}
				//console.log(tris_right_edges);
				var edges_info = [];
				for(let j=0;j<tris_right_edges.length;j++)
				{
					let edge_index = triangles[tris_right_edges[j].tri_index].edges[tris_right_edges[j].edge_index].index;
					if(edges[edge_index].first == i)
						edges_info.push({'q': quartic_cp[edge_index], 'a0':edges[edge_index].alpha_left, 'a1': edges[edge_index].alpha_right, 'b0':edges[edge_index].beta_left, 'b1': edges[edge_index].beta_right});
					else
						edges_info.push({'q': quartic_cp[edge_index].slice().reverse(), 'a0':1-edges[edge_index].alpha_right, 'a1':1-edges[edge_index].alpha_left, 'b0':1-edges[edge_index].beta_right, 'b1':1-edges[edge_index].beta_left});
				}
//				console.log(edges_info);
				var s = new Array(edges_info.length);
				for(let k=0;k<edges_info.length;k++)
				{
					s[k] = 
					vec_sub(
					vec_add(num_vec_mul(edges_info[(k+1)%s.length].b0,edges_info[(k+1)%s.length].q[1]), num_vec_mul(1-edges_info[(k+1)%s.length].b0,edges_info[(k+1)%s.length].q[2])),
					num_vec_mul(1/3,
					vec_sub(
					vec_add(num_vec_mul(edges_info[(k+1)%s.length].a1,edges_info[(k+2)%s.length].q[1]), num_vec_mul(1-edges_info[(k+1)%s.length].a1,edges_info[k].q[1])),
					vec_add(num_vec_mul(edges_info[(k+1)%s.length].b1,edges_info[k].q[0]), num_vec_mul(1-edges_info[(k+1)%s.length].b1,edges_info[(k+1)%s.length].q[1]))
					)));
				}
			//	console.log(s);
				
								
				var m = new Array(s.length);
				for(let k=0;k<s.length;k++)
				{				
					m[k]= new Array(s.length);
					for(let j=0;j<s.length;j++)
					{
						if(k == j)
							m[k][j] = 1- edges_info[(k+1)%s.length].a0;
						else if( (k+1)%s.length == j)
							m[k][j] = edges_info[(k+1)%s.length].a0;
						else
							m[k][j] = 0;
						
					}
				}
				m= new Matrix.Matrix(m);
				var inv = Matrix.inverse(m);
				var t = inv.mmul(new Matrix.Matrix(s.map(r=>[r.x,r.y,r.z])));					
	/*			
				
				console.log(edges_info[0].q[0]);
				for(let k=0;k<edges_info.length;k++)
				{
					console.log(edges_info[k].q[4]);
				}
				console.log(t.data); */
				
				for(let j=0;j<tris_right_edges.length;j++)
				{
					triangles[tris_right_edges[j].tri_index].q[tris_right_edges[j].edge_index][2] = {'x': t.data[j][0],'y':t.data[j][1],'z':t.data[j][2]};
				}
				
			}
			else
			{
				const max_groups_count = triangles.filter(t=>t.splitting == true && t.edges.some(e=>edges[e.index].first == i || edges[e.index].second == i)).length;			
				const left_edge_i = triangles[first_split_i].edges.findIndex(e=>(edges[e.index].first == i && e.dir ==-1) || (edges[e.index].second == i && e.dir ==1));
				var neighbor_info = GetNeighborTri(triangles,first_split_i,left_edge_i);
				var groups = [[{'tri_index': first_split_i, 'edge_index': left_edge_i}]];
				//console.log("max_size",max_groups_count);
				while(groups.length <= max_groups_count)
				{					
					if(triangles[neighbor_info.tri_index].splitting == false)
						// za vsi4ki bez posledniq triugulnik zapisvame lqvoto rebro
						groups[groups.length-1].push({'tri_index': neighbor_info.tri_index, 'edge_index': (neighbor_info.edge_index+2)%3});			
					else
					{ // za posledniq triugulnik ot vsqka grupa zapisvame dqsnoto rebro
						groups[groups.length-1].push({'tri_index': neighbor_info.tri_index, 'edge_index': neighbor_info.edge_index});
						if(neighbor_info.tri_index != first_split_i)
						{ // zapo4vame nova grupa
							groups.push([{'tri_index': neighbor_info.tri_index, 'edge_index': (neighbor_info.edge_index+2)%3}]);
						}
						else
							break;
					}
					neighbor_info = GetNeighborTri(triangles,neighbor_info.tri_index,(neighbor_info.edge_index+2)%3);							
				}
//				console.log(groups);
				groups.forEach(function(g){
					if(g.length == 2)
					{
						if(triangles[g[0].tri_index].z[2*g[0].edge_index] !== undefined && triangles[g[0].tri_index].z[2*g[0].edge_index+1] !== undefined  )	
						{ }
						else
						{
							const Z = FindZVlachkova(triangles,cubic_cp,quartic_cp,g[0].tri_index,g[0].edge_index,g[1]);
							triangles[g[0].tri_index].z[2*g[0].edge_index] = Z.z1;
							triangles[g[0].tri_index].z[2*g[0].edge_index+1] = Z.z2;
							if(g[1].tri_index!=-1)
							{  
								triangles[g[1].tri_index].z[2*g[1].edge_index] = Z.nz2;
								triangles[g[1].tri_index].z[2*g[1].edge_index+1] = Z.nz1;
							}
						}
					}
					else
					{
						var edges_info = [{'q': [triangles[g[0].tri_index].q[(g[0].edge_index+1)%3][0],triangles[g[0].tri_index].q[(g[0].edge_index+1)%3][1] ] }];
						for(let j=0;j<g.length-1;j++)
						{
							let edge_index = triangles[g[j].tri_index].edges[g[j].edge_index].index;
							if(edges[edge_index].first == i)
								edges_info.push({'q': quartic_cp[edge_index], 'a0':edges[edge_index].alpha_left, 'a1': edges[edge_index].alpha_right, 'b0':edges[edge_index].beta_left, 'b1': edges[edge_index].beta_right});
							else
								edges_info.push({'q': quartic_cp[edge_index].slice().reverse(), 'a0':1-edges[edge_index].alpha_right, 'a1':1-edges[edge_index].alpha_left, 'b0':1-edges[edge_index].beta_right, 'b1':1-edges[edge_index].beta_left});
						}
						edges_info.push({'q': [triangles[g[g.length-1].tri_index].q[g[g.length-1].edge_index][0],triangles[g[g.length-1].tri_index].q[g[g.length-1].edge_index][1] ] });
//						console.log(edges_info);
						var s = new Array(edges_info.length-2);
						for(let k=0;k<edges_info.length-2;k++)
						{							
							s[k] = 
							vec_sub(
							vec_add(num_vec_mul(edges_info[k+1].b0,edges_info[k+1].q[1]), num_vec_mul(1-edges_info[k+1].b0,edges_info[k+1].q[2])),
							num_vec_mul(1/3,
							vec_sub(
							vec_add(num_vec_mul(edges_info[k+1].a1,edges_info[k+2].q[1]), num_vec_mul(1-edges_info[k+1].a1,edges_info[k].q[1])),
							vec_add(num_vec_mul(edges_info[k+1].b1,edges_info[k].q[0]), num_vec_mul(1-edges_info[k+1].b1,edges_info[k+1].q[1]))
							)));

						}
						//console.log(s);
						const last_t = FindZVlachkova(triangles,cubic_cp,quartic_cp,g[g.length-1].tri_index,g[g.length-1].edge_index,{'tri_index':-1}).z1;
						// vuvejdame stojnostite na Z
						// purvo za posledniq triugulnik
						triangles[g[g.length-1].tri_index].z[2*g[g.length-1].edge_index ] = last_t;
						//console.log(last_t);
						
						// posle re6avame sistemata za ostanalite
						var t = new Array(s.length+1);
						var sys = [];
						for(let j=0;j<s.length-1;j++)
						{
							let eq_x = (new Array(3*s.length+1)).fill(0);
							let eq_y = (new Array(3*s.length+1)).fill(0);
							let eq_z = (new Array(3*s.length+1)).fill(0);
							eq_x[3*j] = 1-edges_info[j+1].a0;
							eq_x[3*j+3] = edges_info[j+1].a0;
							eq_x[eq_x.length-1] = s[j].x;
							
							eq_y[3*j+1] = 1-edges_info[j+1].a0;
							eq_y[3*j+4] = edges_info[j+1].a0;
							eq_y[eq_y.length-1] = s[j].y;
							
							eq_z[3*j+2] = 1-edges_info[j+1].a0;
							eq_z[3*j+5] = edges_info[j+1].a0;
							eq_z[eq_z.length-1] = s[j].z;
							sys.push(eq_x,eq_y,eq_z);
						}
						let eq_x = (new Array(3*s.length+1)).fill(0);
						let eq_y = (new Array(3*s.length+1)).fill(0);
						let eq_z = (new Array(3*s.length+1)).fill(0);
						var last_t_mul_last_s = num_vec_mul(edges_info[s.length].a0,last_t);
					
						eq_x[eq_x.length-4] = 1-edges_info[s.length].a0;						
						eq_x[eq_x.length-1] = s[s.length-1].x - last_t_mul_last_s.x;
						
						eq_y[eq_y.length-3] = 1-edges_info[s.length].a0;						
						eq_y[eq_y.length-1] = s[s.length-1].y - last_t_mul_last_s.y;
						
						eq_z[eq_z.length-2] = 1-edges_info[s.length].a0;						
						eq_z[eq_z.length-1] = s[s.length-1].z - last_t_mul_last_s.z;
						
						sys.push(eq_x,eq_y,eq_z);
						
						var t = gauss(sys);
//						console.log(sys,t);
						triangles[g[0].tri_index].z[2*g[0].edge_index + 1] = {'x': t[0],'y':t[1],'z':t[2]};
						for(let j=1;j<g.length-1;j++)
						{
							triangles[g[j].tri_index].q[(g[j].edge_index+1)%3][2] = {'x': t[3*j],'y':t[3*j+1],'z':t[3*j+2]};
						}
//////////////// 6te probvame nov metod za minimizirane na oscilaciqta na z
						var null_div = false;
						var z = [];
						for(let j=1;j<edges_info.length-1;j++)
						{
							if(isZero(1-edges_info[j].a0))
							{
								null_div = true;
								break;
							}
						}
						if(!null_div)
						{						
							var Z_vecs = (new Array(edges_info.length-2)).fill(new Object({'x':0,'y':0}));
							let last_i = Z_vecs.length-1;
							Z_vecs[last_i].x = s[last_i].z/(1-edges_info[last_i+1].a0);
							Z_vecs[last_i].y = -edges_info[last_i+1].a0/(1-edges_info[last_i+1].a0);						
							while(--last_i>=0)
							{
								Z_vecs[last_i] = num_vec_mul(-edges_info[last_i+1].a0/(1-edges_info[last_i+1].a0),Z_vecs[last_i+1]);
								Z_vecs[last_i].x = Z_vecs[last_i].x + s[last_i].z/(1-edges_info[last_i+1].a0)
							}
							Z_vecs.push({'x':0,'y':1});
							var sum = {'x':0,'y':0};						
							for(var j=Z_vecs.length-1;j>0;j--)
							{
								var tmp = vec_sub(Z_vecs[j],Z_vecs[j-1]);
								sum = vec_add(sum,num_vec_mul(tmp.y,tmp));
							}
							var last_z = -sum.x/sum.y;
							for(var j=0;j<Z_vecs.length-1;j++)
							{
								z.push(Z_vecs[j].x + Z_vecs[j].y*last_z);
							}
							triangles[g[0].tri_index].z[2*g[0].edge_index + 1].z = z[0];
							for(let j=1;j<g.length-1;j++)
							{
								triangles[g[j].tri_index].q[(g[j].edge_index+1)%3][2].z = z[j];
							}
							// popravqme z na posledniq
							triangles[g[g.length-1].tri_index].z[2*g[g.length-1].edge_index ].z = last_z;
						}
					}
				});
			}

		}
		else
		{ // grani4na to4ka
			// namirame naj-lvoto i naj-dqsnoto rebro izliza6to ot vurha
			//console.log("grani4na",i);
			var right_edge = {'tri_index': -1, 'edge_index': -1};
			for(let t=0;t<triangles.length;t++)
			{ 
				var ready = false;
				for(let e_i = 0;e_i<3;e_i++)
				{
					if(((triangles[t].edges[e_i].dir == 1 && edges[triangles[t].edges[e_i].index].first == i) || (triangles[t].edges[e_i].dir == -1 && edges[triangles[t].edges[e_i].index].second == i))
						&& (GetNeighborTri(triangles,t,e_i).tri_index === -1))
					{
						right_edge.tri_index = t;
						right_edge.edge_index = e_i;
						ready = true;						
						break;
					}
				}
				if(ready)
					break;
			}
			//console.log("desnoto",right_edge);
			//vsi4ki rebra v grupite 6te sa desni
			right_edge.split = false;
			var groups = [[right_edge]];		
			// ako naj-desniq triugulnik e splitnat smqtame z1 za dqsnata polovinka i zapo4vame grupite ot lqvata polovinka
			if(triangles[right_edge.tri_index].splitting == true)
			{// console.log("v 1");
				triangles[right_edge.tri_index].z[2*right_edge.edge_index]=FindZVlachkova(triangles,cubic_cp,quartic_cp,right_edge.tri_index,right_edge.edge_index,{'tri_index':-1}).z1;
				groups[0][0].split = true;
			}
			var neighbor_info = GetNeighborTri(triangles,right_edge.tri_index,(right_edge.edge_index+2)%3);										
			while(neighbor_info.tri_index != -1)
			{					
				if(triangles[neighbor_info.tri_index].splitting == false)		
					groups[groups.length-1].push({'tri_index': neighbor_info.tri_index, 'edge_index': neighbor_info.edge_index, 'split':false});			
				else
				{ 
					groups[groups.length-1].push({'tri_index': neighbor_info.tri_index, 'edge_index': neighbor_info.edge_index, 'split':true});
					groups.push([{'tri_index': neighbor_info.tri_index, 'edge_index': neighbor_info.edge_index, 'split':true}]);
				}
				neighbor_info = GetNeighborTri(triangles,neighbor_info.tri_index,(neighbor_info.edge_index+2)%3);							
			}
			//console.log(groups);
			if(groups[groups.length-1].length == 1)
			{ // ako poslednata grupa ima samo 1 element zna4i tova e polovinkata na splitnat triugulnik i za nego smqtame z2 kato za grani4no rebro
				triangles[groups[groups.length-1][0].tri_index].z[2*((groups[groups.length-1][0].edge_index+2)%3)+1]=FindZVlachkova(triangles,cubic_cp,quartic_cp,groups[groups.length-1][0].tri_index,(groups[groups.length-1][0].edge_index+2)%3,{'tri_index':-1}).z2;			
				groups.splice(-1);
			}
			groups.forEach(function(g){
				if(g.length == 2 && g[0].split == true && g[1].split == true)
				{	// dva triugulnika za splitvane edin do drug smqtame direktno z1 i z2
					if(triangles[g[0].tri_index].z[2*g[0].edge_index] !== undefined && triangles[g[0].tri_index].z[2*g[0].edge_index+1] !== undefined  )	
					{ }
					else
					{ 
						const Z = FindZVlachkova(triangles,cubic_cp,quartic_cp,g[0].tri_index,(g[0].edge_index+2)%3,g[1]);						
						triangles[g[0].tri_index].z[2*((g[0].edge_index+2)%3)] = Z.z1;
						triangles[g[0].tri_index].z[2*((g[0].edge_index+2)%3)+1] = Z.z2;		
						if(g[1].tri_index!=-1)
						{  
							triangles[g[1].tri_index].z[2*g[1].edge_index] = Z.nz2;
							triangles[g[1].tri_index].z[2*g[1].edge_index+1] = Z.nz1;
						}
					}
					
				}
				else
				{
					var edges_info = [];
					for(let j=0;j<g.length;j++)
					{
						let edge_index = triangles[g[j].tri_index].edges[g[j].edge_index].index;
						if(j == 0 )
						{
							if(g[j].split === true)
								edges_info.push({'q': [triangles[g[0].tri_index].q[g[0].edge_index][0],triangles[g[0].tri_index].q[g[0].edge_index][1] ] });
							else
							{
								if(edges[edge_index].first == i)
									edges_info.push({'q': quartic_cp[edge_index],});
								else
									edges_info.push({'q': quartic_cp[edge_index].slice().reverse()});			
							}								
						}
						else if(edges[edge_index].first == i)
							edges_info.push({'q': quartic_cp[edge_index], 'a0':edges[edge_index].alpha_left, 'a1': edges[edge_index].alpha_right, 'b0':edges[edge_index].beta_left, 'b1': edges[edge_index].beta_right});
						else
							edges_info.push({'q': quartic_cp[edge_index].slice().reverse(), 'a0':1-edges[edge_index].alpha_right, 'a1':1-edges[edge_index].alpha_left, 'b0':1-edges[edge_index].beta_right, 'b1':1-edges[edge_index].beta_left});


					}
					if(g[g.length-1].split === true)
						edges_info.push({'q': [triangles[g[g.length-1].tri_index].q[g[g.length-1].edge_index][0],triangles[g[g.length-1].tri_index].q[g[g.length-1].edge_index][1] ] });
					else
					{
						let edge_index = triangles[g[g.length-1].tri_index].edges[(g[g.length-1].edge_index+2)%3].index;
						if(edges[edge_index].first == i)
							edges_info.push({'q': quartic_cp[edge_index]});
						else
							edges_info.push({'q': quartic_cp[edge_index].slice().reverse()});
						
					}
//					console.log(edges_info);
					var s = new Array(edges_info.length-2);
					for(let k=0;k<edges_info.length-2;k++)
					{							
						s[k] = 
						vec_add(
						vec_add(num_vec_mul(edges_info[k+1].b0,edges_info[k+1].q[1]), num_vec_mul(1-edges_info[k+1].b0,edges_info[k+1].q[2])),
						num_vec_mul(-1/3,
							vec_add(
							vec_add(num_vec_mul(edges_info[k+1].a1,edges_info[k+2].q[1]), num_vec_mul(1-edges_info[k+1].a1,edges_info[k].q[1])),
							vec_add(num_vec_mul(-edges_info[k+1].b1,edges_info[k].q[0]), num_vec_mul(-1+edges_info[k+1].b1,edges_info[k+1].q[1]))
						)));
					}
					var last_t;
					if(g[g.length-1].split == true)
					{  
						last_t = FindZVlachkova(triangles,cubic_cp,quartic_cp,g[g.length-1].tri_index,g[g.length-1].edge_index,{'tri_index':-1}).z1;
						triangles[g[g.length-1].tri_index].z[2*g[g.length-1].edge_index] = last_t;
					}
					else
					{
						last_t = FindZVlachkova(triangles,cubic_cp,quartic_cp,g[g.length-1].tri_index,g[g.length-1].edge_index,{'tri_index':-1}).z1;
						triangles[g[g.length-1].tri_index].q[g[g.length-1].edge_index][2] = last_t;
					}
//					console.log(last_t);
					// posle re6avame sistemata za ostanalite
					var t = new Array(s.length+1);
					var sys = [];
					for(let j=0;j<s.length-1;j++)
					{
						let eq_x = (new Array(3*s.length+1)).fill(0);
						let eq_y = (new Array(3*s.length+1)).fill(0);
						let eq_z = (new Array(3*s.length+1)).fill(0);
						eq_x[3*j] = 1-edges_info[j+1].a0;
						eq_x[3*j+3] = edges_info[j+1].a0;
						eq_x[eq_x.length-1] = s[j].x;
						
						eq_y[3*j+1] = 1-edges_info[j+1].a0;
						eq_y[3*j+4] = edges_info[j+1].a0;
						eq_y[eq_y.length-1] = s[j].y;
						
						eq_z[3*j+2] = 1-edges_info[j+1].a0;
						eq_z[3*j+5] = edges_info[j+1].a0;
						eq_z[eq_z.length-1] = s[j].z;
						sys.push(eq_x,eq_y,eq_z);
					}
					let eq_x = (new Array(3*s.length+1)).fill(0);
					let eq_y = (new Array(3*s.length+1)).fill(0);
					let eq_z = (new Array(3*s.length+1)).fill(0);
					var last_t_mul_last_s = num_vec_mul(edges_info[s.length].a0,last_t);
				
					eq_x[eq_x.length-4] = 1-edges_info[s.length].a0;						
					eq_x[eq_x.length-1] = s[s.length-1].x - last_t_mul_last_s.x;
					
					eq_y[eq_y.length-3] = 1-edges_info[s.length].a0;						
					eq_y[eq_y.length-1] = s[s.length-1].y - last_t_mul_last_s.y;
					
					eq_z[eq_z.length-2] = 1-edges_info[s.length].a0;						
					eq_z[eq_z.length-1] = s[s.length-1].z - last_t_mul_last_s.z;
					
					sys.push(eq_x,eq_y,eq_z);
					var t = gauss(sys);
//					console.log(t);
					for(let j=0;j<g.length-1;j++)
					{ 
						if(g[j].split === true)
						{
							triangles[g[j].tri_index].z[2*((g[j].edge_index+2)%3)+1]  = {'x': t[3*j],'y':t[3*j+1],'z':t[3*j+2]};
						}
						else
							triangles[g[j].tri_index].q[g[j].edge_index][2] = {'x': t[3*j],'y':t[3*j+1],'z':t[3*j+2]};
											
					}
//////////////// 6te probvame nov metod za minimizirane na oscilaciqta na z
					var null_div = false;
					var z = [];
					for(let j=1;j<edges_info.length-1;j++)
					{
						if(isZero(1-edges_info[j].a0))
						{
							null_div = true;
							break;
						}
					}
					if(!null_div)
					{
						var Z_vecs = (new Array(edges_info.length-2)).fill(new Object({'x':0,'y':0}));
						let last_i = Z_vecs.length-1;
						Z_vecs[last_i].x = s[last_i].z/(1-edges_info[last_i+1].a0);
						Z_vecs[last_i].y = -edges_info[last_i+1].a0/(1-edges_info[last_i+1].a0);						
						while(--last_i>=0)
						{
							Z_vecs[last_i] = num_vec_mul(-edges_info[last_i+1].a0/(1-edges_info[last_i+1].a0),Z_vecs[last_i+1]);
							Z_vecs[last_i].x = Z_vecs[last_i].x + s[last_i].z/(1-edges_info[last_i+1].a0)
						}
						Z_vecs.push({'x':0,'y':1});
						var sum = {'x':0,'y':0};
						for(var j=Z_vecs.length-1;j>0;j--)
						{
							var tmp = vec_sub(Z_vecs[j],Z_vecs[j-1]);
							sum = vec_add(sum,num_vec_mul(tmp.y,tmp));
						}
						var last_z = -sum.x/sum.y;
						for(var j=0;j<Z_vecs.length-1;j++)
						{
							z.push(Z_vecs[j].x + Z_vecs[j].y*last_z);
						}
						for(let j=0;j<g.length-1;j++)
						{ 
							if(g[j].split === true)
							{
								triangles[g[j].tri_index].z[2*((g[j].edge_index+2)%3)+1].z  = z[j];
							}
							else
								triangles[g[j].tri_index].q[g[j].edge_index][2].z = z[j];
												
						}
						// poslednata to4ka
						if(g[g.length-1].split == true)
						{  
							triangles[g[g.length-1].tri_index].z[2*g[g.length-1].edge_index].z = last_z;
						}
						else
						{
							triangles[g[g.length-1].tri_index].q[g[g.length-1].edge_index][2].z = last_z;
						}
					}
				}
			});
			
		}
	}
//	console.log(triangles);
}

function Splitting(triangles,cubic_cp,quartic_cp)
{
	for(var i=0;i<triangles.length;i++)
	{
		// Find Zi
		for(var j=0;j<3;j++)
		{  
			if(triangles[i].z[2*j] !== undefined && triangles[i].z[2*j+1] !== undefined  )	
				continue;
			const neighbor_info = GetNeighborTri(triangles,i,j);
			const Z = FindZVlachkova(triangles,cubic_cp,quartic_cp,i,j,neighbor_info);
			triangles[i].z[2*j] = Z.z1;
			triangles[i].z[2*j+1] = Z.z2;
			if(neighbor_info.tri_index!=-1)
			{  
				triangles[neighbor_info.tri_index].z[2*neighbor_info.edge_index] = Z.nz2;
				triangles[neighbor_info.tri_index].z[2*neighbor_info.edge_index+1] = Z.nz1;
			}
		}
		// 3
		triangles[i].q[0][2] = num_vec_mul(1/3, vec_add(triangles[i].q[0][1], vec_add(triangles[i].z[0],triangles[i].z[5])) );
		triangles[i].q[1][2] = num_vec_mul(1/3, vec_add(triangles[i].q[1][1], vec_add(triangles[i].z[1],triangles[i].z[2])) );
		triangles[i].q[2][2] = num_vec_mul(1/3, vec_add(triangles[i].q[2][1], vec_add(triangles[i].z[3],triangles[i].z[4])) );
		// 4
		const p1 = vec_add(vec_sub(num_vec_mul(2,triangles[i].q[0][2]),num_vec_mul(4/3,triangles[i].q[0][1])),num_vec_mul(1/3,triangles[i].q[0][0])); 
		const p2 = vec_add(vec_sub(num_vec_mul(2,triangles[i].q[1][2]),num_vec_mul(4/3,triangles[i].q[1][1])),num_vec_mul(1/3,triangles[i].q[1][0])); 
		const p3 = vec_add(vec_sub(num_vec_mul(2,triangles[i].q[2][2]),num_vec_mul(4/3,triangles[i].q[2][1])),num_vec_mul(1/3,triangles[i].q[2][0])); 
		// 5
		triangles[i].z[6] = num_vec_mul(1/3,vec_add(vec_add(p1,p2),p3));
		//6
		triangles[i].q[0][3] = vec_add(num_vec_mul(0.75,p1),num_vec_mul(0.25,triangles[i].z[6]));
		triangles[i].q[1][3] = vec_add(num_vec_mul(0.75,p2),num_vec_mul(0.25,triangles[i].z[6]));
		triangles[i].q[2][3] = vec_add(num_vec_mul(0.75,p3),num_vec_mul(0.25,triangles[i].z[6]));
		//7
		triangles[i].x[0] = vec_sub(num_vec_mul(3/2,vec_sub(vec_add(triangles[i].q[1][3],triangles[i].q[2][3]),triangles[i].q[0][3])),num_vec_mul(1/2,vec_sub(vec_add(triangles[i].q[1][2],triangles[i].q[2][2]),triangles[i].q[0][2])));
		triangles[i].x[1] = vec_sub(num_vec_mul(3/2,vec_sub(vec_add(triangles[i].q[0][3],triangles[i].q[2][3]),triangles[i].q[1][3])),num_vec_mul(1/2,vec_sub(vec_add(triangles[i].q[0][2],triangles[i].q[2][2]),triangles[i].q[1][2])));
		triangles[i].x[2] = vec_sub(num_vec_mul(3/2,vec_sub(vec_add(triangles[i].q[1][3],triangles[i].q[0][3]),triangles[i].q[2][3])),num_vec_mul(1/2,vec_sub(vec_add(triangles[i].q[1][2],triangles[i].q[0][2]),triangles[i].q[2][2])));
	}
}

function SplittingVlachkova(triangles)
{
	triangles.filter(t=>t.splitting == true).forEach(function(t){
		// 3
		t.q[0][2] = num_vec_mul(1/3, vec_add(t.q[0][1], vec_add(t.z[0],t.z[5])) );
		t.q[1][2] = num_vec_mul(1/3, vec_add(t.q[1][1], vec_add(t.z[1],t.z[2])) );
		t.q[2][2] = num_vec_mul(1/3, vec_add(t.q[2][1], vec_add(t.z[3],t.z[4])) );
		// 4
		t.x[0] = num_vec_mul(0.5,vec_add(t.q[1][2],t.q[2][2]));
		t.x[1] = num_vec_mul(0.5,vec_add(t.q[0][2],t.q[2][2]));
		t.x[2] = num_vec_mul(0.5,vec_add(t.q[0][2],t.q[1][2]));			
		//5
		t.q[0][3] = num_vec_mul(1/3, vec_add(t.q[0][2],vec_add(t.x[1],t.x[2])));
		t.q[1][3] = num_vec_mul(1/3, vec_add(t.q[1][2],vec_add(t.x[0],t.x[2])));
		t.q[2][3] = num_vec_mul(1/3, vec_add(t.q[2][2],vec_add(t.x[0],t.x[1])));		
		//6
		t.z[6] = num_vec_mul(1/3,vec_add(vec_add(t.q[0][3],t.q[1][3]),t.q[2][3]));
	});

}

function DiagonalCoef(a,b,c,d)
{ 
	// ac Z bd = a + result.alpha*ac
	// ac Z bd = b + result.beta*bd
	const ab = vec_sub(b,a);
	const ac = vec_sub(c,a);
	const db = vec_sub(b,d);	
	const alpha_beta = gauss([[ac.x,db.x,0,ab.x],[ac.y,db.y,0,ab.y],[ac.z,db.z,0,ab.z]]);
	return {"alpha" : alpha_beta[0], "beta": alpha_beta[1]};
}

function FindZVlachkova(triangles,cubic_cp,quartic_cp,i,j,neighbor_info)
{
//	console.log(triangles[i],i);
	const q10 = triangles[i].q[j][0];
	const q20 = triangles[i].q[(j+1)%3][0];
	var q11 = triangles[i].q[j][1];
	var q21 = triangles[i].q[(j+1)%3][1];
	if(neighbor_info.tri_index!=-1)
	{ 
		const P  = triangles[i].edges[j].dir > 0 ? quartic_cp[triangles[i].edges[j].index] : quartic_cp[triangles[i].edges[j].index].slice().reverse();				
		var nq11 = triangles[neighbor_info.tri_index].q[(neighbor_info.edge_index+1)%3][1];
		var nq21 = triangles[neighbor_info.tri_index].q[neighbor_info.edge_index][1];
		if(triangles[i].edges[j].dir == triangles[neighbor_info.tri_index].edges[neighbor_info.edge_index].dir)
		{
			var tmp = nq11; nq1= nq21; nq21 = tmp;
			console.log("bad triangulation array");
		}
		const a = DiagonalCoef(P[1],q11,q10,nq11).alpha;
		const b = DiagonalCoef(q20,q21,P[3],nq21).alpha;
		//console.log(a,b);
		
		const m1 = vec_add(num_vec_mul(a,q10), num_vec_mul(1-a,P[1]));
		const m2 = vec_add(num_vec_mul(b,P[3]), num_vec_mul(1-b,q20));
		
		const c1 = vec_add(num_vec_mul(a,P[1]), num_vec_mul(1-a,P[2]));
		const c2 = vec_add(num_vec_mul(b,P[2]), num_vec_mul(1-b,P[3]));
	
		const z11 = vec_add(q11,vec_sub(c1,m1));
		const z12 = vec_add(q11,vec_sub(c2,m1));
		const nz11 = vec_add(nq11,vec_sub(c1,m1));
		const nz12 = vec_add(nq11,vec_sub(c2,m1));
		const z21 = vec_add(q21,vec_sub(c1,m2));
		const z22 = vec_add(q21,vec_sub(c2,m2));
		const nz21 = vec_add(nq21,vec_sub(c1,m2));
		const nz22 = vec_add(nq21,vec_sub(c2,m2));
		
		return {"z1" :  vec_add(num_vec_mul(2/3,z11),num_vec_mul(1/3,z21)),
				"nz1" : vec_add(num_vec_mul(2/3,nz11),num_vec_mul(1/3,nz21)),
				"z2" :  vec_add(num_vec_mul(1/3,z12),num_vec_mul(2/3,z22)),
				"nz2" : vec_add(num_vec_mul(1/3,nz12),num_vec_mul(2/3,nz22))};
		 
	}
	else
	{ 
		
		if(triangles[i].splitting !== undefined && triangles[i].splitting === false)
		{  
			q11 = triangles[i].edges[(j+2)%3].dir > 0 ? quartic_cp[triangles[i].edges[(j+2)%3].index][3] : quartic_cp[triangles[i].edges[(j+2)%3].index][1];
			q21 = triangles[i].edges[(j+1)%3].dir < 0 ?  quartic_cp[triangles[i].edges[(j+1)%3].index][3] : quartic_cp[triangles[i].edges[(j+1)%3].index][1];		
		}		
		const C  = triangles[i].edges[j].dir > 0 ? cubic_cp[triangles[i].edges[j].index] : cubic_cp[triangles[i].edges[j].index].slice().reverse();							
		const r11 = vec_add(vec_sub(q11,q10) ,C[1]);
		const r12 = vec_add(vec_sub(q21,q20) ,C[1]);
		const r21 = vec_add(vec_sub(q11,q10) ,C[2]);
		const r22 = vec_add(vec_sub(q21,q20) ,C[2]);
		return {"z2" : vec_add(num_vec_mul(1/3,r21),num_vec_mul(2/3,r22)),
				"z1" : vec_add(num_vec_mul(2/3,r11),num_vec_mul(1/3,r12)) };

	}
}

function FindZChiyokura(triangles,i,j,cubic_cp)
{
	const q1 = triangles[i].q[j][1];
	const q2 = triangles[i].q[(j+1)%3][1];
	const C  = triangles[i].edges[j].dir < 0 ? cubic_cp[triangles[i].edges[j].index] : cubic_cp[triangles[i].edges[j].index].slice().reverse();			
	const neighbor_info = GetNeighborTri(triangles,i,j);
	if(neighbor_info.tri_index!=-1)
	{
		//console.log(i,j,triangles[i].edges[j],triangles[neighbor_info.tri_index].edges[neighbor_info.edge_index]);
		//console.log(triangles[i].q[j][0],triangles[neighbor_info.tri_index].q[(neighbor_info.edge_index+1)%3][0]);
		var nq1 = triangles[neighbor_info.tri_index].q[(neighbor_info.edge_index+1)%3][1];
		var nq2 = triangles[neighbor_info.tri_index].q[neighbor_info.edge_index][1];
		if(triangles[i].edges[j].dir == triangles[neighbor_info.tri_index].edges[neighbor_info.edge_index].dir)
		{
			var tmp = nq1; nq1= nq2; nq2 = tmp;
			console.log("bad triangulation array");
		}
		const nq1_q1 = vec_sub(q1,nq1);
		const nq2_q2 = vec_sub(q2,nq2);
		const b3c3 =  num_vec_mul(1/len(nq1_q1),nq1_q1);
		const b0c0 =  num_vec_mul(1/len(nq2_q2),nq2_q2);
		
		
		const c0c1 = vec_sub(C[1],C[0]);				
		const c0q2 = vec_sub(q2,C[0]);
		const k0_h0 = gauss([[c0c1.x,b0c0.x,0,c0q2.x],[c0c1.y,b0c0.y,0,c0q2.y],[c0c1.z,b0c0.z,0,c0q2.z]]);
		const k0 = k0_h0[0];
		const h0 = k0_h0[1];
				
		const c2c3 = vec_sub(C[3],C[2]);				
		const c3q1 = vec_sub(q1,C[3]);
		const k1_h1 = gauss([[c2c3.x,b3c3.x,0,c3q1.x],[c2c3.y,b3c3.y,0,c3q1.y],[c2c3.z,b3c3.z,0,c3q1.z]]);
		const k1 = k1_h1[0];
		const h1 = k1_h1[1];

		return {"z2" : vec_add(C[1],num_vec_mul(1/3, vec_add( num_vec_mul(k0+k1,b0c0), vec_add( num_vec_mul(k0,b3c3), vec_add( num_vec_mul(2*h0,vec_sub(C[2],C[1])), num_vec_mul(h1,c0c1) ))) )),
				"z1" : vec_add(C[2],num_vec_mul(1/3,vec_add(num_vec_mul(k1,b0c0), vec_add(num_vec_mul(k0+k1,b3c3),vec_add(num_vec_mul(h0,c2c3),num_vec_mul(2*h1,vec_sub(C[2],C[1]))))) )) };
		 
	}
	else
	{
		const r11 = vec_add(vec_sub(q1,triangles[i].q[j][0]) ,C[1]);
		const r12 = vec_add(vec_sub(q2,triangles[i].q[(j+1)%3][0]) ,C[1]);
		const r21 = vec_add(vec_sub(q1,triangles[i].q[j][0]) ,C[2]);
		const r22 = vec_add(vec_sub(q2,triangles[i].q[(j+1)%3][0]) ,C[2]);
		return {"z2" : vec_add(num_vec_mul(1/3,r21),num_vec_mul(2/3,r22)),
				"z1" : vec_add(num_vec_mul(2/3,r11),num_vec_mul(1/3,r12)) };

	}
}

function GetNeighborTri(triangles,tri_index,edge_index)
{
	for(var i=0;i<triangles.length;i++)
	{
		if(i!=tri_index)
		{
			for(var j=0;j<3;j++)
			{ 
				if(triangles[i].edges[j].index == triangles[tri_index].edges[edge_index].index)
					return {"tri_index" : i , "edge_index" : j};
			}
		}
	}
	return {"tri_index" : -1 , "edge_index" : -1};
}

function Bezier3To4(control_polygons)
{
	var new_control_polygons = [];
	control_polygons.forEach(function(cp){
		new_control_polygons.push(new Array());
		const last_i = new_control_polygons.length-1;
		new_control_polygons[last_i].push(cp[0]);
		new_control_polygons[last_i].push({"x": 0.25*cp[0].x + 0.75*cp[1].x,"y": 0.25*cp[0].y + 0.75*cp[1].y, "z": 0.25*cp[0].z + 0.75*cp[1].z});
		new_control_polygons[last_i].push({"x": 0.5*(cp[1].x + cp[2].x),"y": 0.5*(cp[1].y + cp[2].y), "z": 0.5*(cp[1].z + cp[2].z)});
		new_control_polygons[last_i].push({"x": 0.75*cp[2].x + 0.25*cp[3].x,"y": 0.75*cp[2].y + 0.25*cp[3].y, "z": 0.75*cp[2].z + 0.25*cp[3].z});
		new_control_polygons[last_i].push(cp[3]);
	});
	return new_control_polygons;
}

function EdgesToBezier3(points,edges)
{
	var control_polygons = [];
	for(var i=0;i<edges.length;i++)
	{
		const l = dist_2(points[edges[i].first],points[edges[i].second]);
		var c = [];
		if(edges[i].h !== undefined)
		{
			var left,right;
			if(isZero(edges[i].h) || (isZero(edges[i].v[0]) && isZero(edges[i].v[1])))
			{
				let a = edges[i].w;
				let h = edges[i].h;
				let k = l-h;
				c.push( a[0]*h*h*h + a[1]*h*h + a[2]*h + a[3]);
				c.push( a[0]*h*h*k + (2/3)*a[1]*k*h + (1/3)*a[2]*k);
				c.push( a[0]*h*k*k + (1/3)*a[1]*k*k );
				c.push( 3*c[1] + 3*c[2] + a[0]*k*k*k + c[0]);
				c[2] = 2*c[1] + c[2] + c[0];
				c[1] = c[1] + c[0];		
				left = vec_add(num_vec_mul(1-h/l,points[edges[i].first]), num_vec_mul(h/l,points[edges[i].second]));
				right = points[edges[i].second];
			}
			else
			{
				let alpha = edges[i].v;
				let h = edges[i].h;
				c.push( alpha[3] );
				c.push( c[0] + alpha[2]*h*(1/3) );
				c.push( c[0] + alpha[2]*h*(2/3) + alpha[1]*h*h*(1/3) );
				c.push( c[0] + alpha[2]*h + alpha[1]*h*h + alpha[0]*h*h*h );
				left =  points[edges[i].first]; 
				right = vec_add(num_vec_mul(1-h/l,points[edges[i].first]), num_vec_mul(h/l,points[edges[i].second]));
				
			}
			
			control_polygons.push(new Array());
			if(isZero(edges[i].h) || isZero(edges[i].h-l))
			{ 
				for(var j=0;j<=3;j++)
				{
					control_polygons[i].push({"x":(1-j/3)*left.x + j/3*right.x ,
											  "y":(1-j/3)*left.y + j/3*right.y,
											  "z":c[j]});
				}	
			}
			else if(isZero(edges[i].v[0]) && isZero(edges[i].v[1]))
			{
				control_polygons[i].push(points[edges[i].first]);
				for(var j=1;j<=3;j++)
				{
					control_polygons[i].push({"x":(1-j/3)*left.x + j/3*right.x ,
											  "y":(1-j/3)*left.y + j/3*right.y,
											  "z":c[j]});
				}	
			}
			else
			{ 
				for(var j=0;j<=2;j++)
				{
					control_polygons[i].push({"x":(1-j/3)*left.x + j/3*right.x ,
											  "y":(1-j/3)*left.y + j/3*right.y,
											  "z":c[j]});
				}	
				control_polygons[i].push(points[edges[i].second]);
			}
		}
		else
		{
			c.push( edges[i].alpha[3] );
			c.push( c[0] + edges[i].alpha[2]*l*(1/3) );
			c.push( c[0] + edges[i].alpha[2]*l*(2/3) + edges[i].alpha[1]*l*l*(1/3) );
			c.push( c[0] + edges[i].alpha[2]*l + edges[i].alpha[1]*l*l + edges[i].alpha[0]*l*l*l );
			control_polygons.push(new Array());
			for(var j=0;j<=3;j++)
			{
				control_polygons[i].push({"x":(1-j/3)*points[edges[i].first].x + j/3*points[edges[i].second].x ,
										  "y":(1-j/3)*points[edges[i].first].y + j/3*points[edges[i].second].y,
										  "z":c[j]});
			}
		}
	}
	return control_polygons;
}

function TrinagleToBezierControlPoints(t,quartic_cp)
{
	var result = [];
	if(t.splitting == true)
	{
		for(var i=0;i<3;i++)
		{
			var control_points = new Map();
			if(t.edges[i].dir > 0)
				for(var j = 0;j<=4;j++)
					control_points.set(j+",0",quartic_cp[t.edges[i].index][j]);
			else
				for(var j = 0;j<=4;j++)
					control_points.set(j+",0",quartic_cp[t.edges[i].index][4-j]);

			control_points.set("0,1",t.q[i][1]);
			control_points.set("1,1",t.z[2*i]);
			control_points.set("2,1",t.z[2*i+1]);
			control_points.set("3,1",t.q[(i+1)%3][1]);

			control_points.set("0,2",t.q[i][2]);
			control_points.set("1,2",t.x[(i+2)%3]);
			control_points.set("2,2",t.q[(i+1)%3][2]);

			control_points.set("0,3",t.q[i][3]);
			control_points.set("1,3",t.q[(i+1)%3][3]);
			
			control_points.set("0,4",t.z[6]);
			result.push(control_points);
		}
	}
	else
	{ 
		var control_points = new Map();
		if(t.edges[0].dir > 0)
			for(var j = 0;j<=4;j++)
				control_points.set(j+",0",quartic_cp[t.edges[0].index][j]);
		else
			for(var j = 0;j<=4;j++)
				control_points.set(j+",0",quartic_cp[t.edges[0].index][4-j]);
		
		if(t.edges[2].dir < 0)
			for(var j = 1;j<=4;j++)
				control_points.set("0,"+j,quartic_cp[t.edges[2].index][j]);
		else
			for(var j = 1;j<=4;j++)
				control_points.set("0,"+j,quartic_cp[t.edges[2].index][4-j]);
			
		if(t.edges[1].dir > 0)
			for(var j = 1;j<4;j++)
				control_points.set(4-j+","+j,quartic_cp[t.edges[1].index][j]);
		else
			for(var j = 1;j<4;j++)
				control_points.set(4-j+","+j,quartic_cp[t.edges[1].index][4-j]);
		
		control_points.set("1,1",t.q[0][2]);
		control_points.set("2,1",t.q[1][2]);
		control_points.set("1,2",t.q[2][2]);
		result.push(control_points);
	}
	return result;
}


function BezierTrianglesMesh(control_points,degree,grid_size=11)
{
	// points[0] = Bezier(0,0,1) =  Bezier(w) 
	// points [grid_size-1] = Bezier(1,0,0) = Bezier(u)
	// points[(grid_size+1)*grid_size - 1] = Bezier(0,1,0) =  Bezier(v)
	//	   v= 9 
	//       7 8
	//      4 5 6  
	//  w= 0 1 2 3 = u
	var points = new Object({'x': [], 'y' : [], 'z' : []}); 
	var edges =new Object({'i': [], 'j' : [], 'k' : []}); 
	for(var index = 0,row = 0, col=0; index < (grid_size*(grid_size+1)*0.5)<<0 ; index++)
	{ 
		if(row + col >= grid_size)
		{
			row++; col = 0;
		}
		const v1 = row/(grid_size-1);
		const w1 = (grid_size-1-row-col)/(grid_size-1);
		const u1 = 1-v1-w1;
		
	//	console.log("row/col",row,col, "value" ,u1,v1,w1);
		
		points.x.push(0);
		points.y.push(0);
		points.z.push(0);
		for(var i = degree; i>=0;i--)
		{
			for(var j= degree - i ; j>=0; j--)
			{ 
				const polynom_coeff = BernsteinCoefficients.get(degree)[j][i]* Math.pow(u1, i)*Math.pow(v1, j)*Math.pow(w1, degree -i - j);
				const cp = control_points.get(i+","+j);
				points.x[index] =  polynom_coeff*cp.x + points.x[index];
				points.y[index] =  polynom_coeff*cp.y + points.y[index];
				points.z[index] =  polynom_coeff*cp.z + points.z[index];
			}
		}
		
		if(row+col+1 < grid_size)
		{
			edges.i.push(index);
			edges.j.push(index+1)
			edges.k.push(index+grid_size-row);
			
			if(row>0)
			{
				edges.i.push(index);
				edges.k.push(index+1)
				edges.j.push(index-grid_size+row);
			}
		}
		
		col++;
	}
	return {'points': points , 'edges' : edges } ;
}

function ControlPointsMapToMesh(control_points,degree)
{
	var points = new Object({'x': [], 'y' : [], 'z' : []}); 
	for(var row= degree;row>0;row--)
	{
		if(row%2 ==0)
		{
			for(var col=0;col+row<=degree;col++)
			{
				points.x.push(control_points.get(col+","+row).x);
				points.y.push(control_points.get(col+","+row).y);
				points.z.push(control_points.get(col+","+row).z);
				
				points.x.push(control_points.get(col+","+(row-1)).x);
				points.y.push(control_points.get(col+","+(row-1)).y);
				points.z.push(control_points.get(col+","+(row-1)).z);
				
				points.x.push(control_points.get((col+1)+","+(row-1)).x);
				points.y.push(control_points.get((col+1)+","+(row-1)).y);
				points.z.push(control_points.get((col+1)+","+(row-1)).z);
				
				points.x.push(control_points.get(col+","+row).x);
				points.y.push(control_points.get(col+","+row).y);
				points.z.push(control_points.get(col+","+row).z);
			}
		}
		else
		{
			for(var col=degree-row;col>=0;col--)
			{
				points.x.push(control_points.get(col+","+row).x);
				points.y.push(control_points.get(col+","+row).y);
				points.z.push(control_points.get(col+","+row).z);
				
				points.x.push(control_points.get(col+","+(row-1)).x);
				points.y.push(control_points.get(col+","+(row-1)).y);
				points.z.push(control_points.get(col+","+(row-1)).z);
				
				points.x.push(control_points.get((col+1)+","+(row-1)).x);
				points.y.push(control_points.get((col+1)+","+(row-1)).y);
				points.z.push(control_points.get((col+1)+","+(row-1)).z);
				
				points.x.push(control_points.get(col+","+row).x);
				points.y.push(control_points.get(col+","+row).y);
				points.z.push(control_points.get(col+","+row).z);
			}
		}
	}
	return points;
}

function BezierCurve(bezier_control_points)
{	
	const degree = bezier_control_points.length - 1;
	var precision = 40 * degree;
	if(degree > 1)
	{
		GeneratePascalTriangle(degree);
		var result = new Array();
		result.push(bezier_control_points[0]);
		const interval_len = 1.0/precision;
		for(var i=0; i < precision ; i++)
		{
			const param = i* interval_len + Math.random() * interval_len;
			result.push(PointOnBezierCurve(bezier_control_points,param));
		}
		result.push(bezier_control_points[degree]);
		return result;
	}
	return bezier_control_points;

}

function PointOnBezierCurve(bezier_control_points, t)
{
	const degree = bezier_control_points.length - 1;
	var result = {"x" : 0.0 ,"y" : 0.0, "z" : 0.0};
	for(var i=0; i<=degree; i++)
	{
		const bernstein_polynom_term =  pascal_triangle[degree][i]* Math.pow(t, i) * Math.pow(1-t, degree-i);
		result.x = result.x + bernstein_polynom_term * bezier_control_points[i].x;
		result.y = result.y + bernstein_polynom_term * bezier_control_points[i].y;
		result.z = result.z + bernstein_polynom_term * bezier_control_points[i].z;
		
	}
	return result;
}

function GeneratePascalTriangle(row_num)
{
	if(pascal_triangle.length > row_num)
		return;
	if(pascal_triangle.length == 0)
		pascal_triangle.push([1]);
	for(var j=pascal_triangle.length ; j<=row_num ; j++)
	{
		pascal_triangle.push(new Array(j+1));
		pascal_triangle[j][0] = 1; pascal_triangle[j][j] = 1;
		for(var k = 1; k < j; k++ )
			pascal_triangle[j][k] = pascal_triangle[j-1][k-1] + pascal_triangle[j-1][k];
	}
}

function vec_add(a,b)
{
	if(a.hasOwnProperty("z") && b.hasOwnProperty("z"))
		return {"x" : a.x + b.x ,"y" : a.y + b.y , "z" : a.z + b.z};
	else
		return {"x" : a.x + b.x ,"y" : a.y + b.y };
}

function vec_sub(a,b)
{
	if(a.hasOwnProperty("z") && b.hasOwnProperty("z"))
		return {"x" : a.x - b.x ,"y" : a.y - b.y , "z" : a.z - b.z};
	else
		return {"x" : a.x - b.x ,"y" : a.y - b.y };
}

function num_vec_mul(num,vec)
{
	if(vec.hasOwnProperty("z"))
		return {"x" : num*vec.x  ,"y" : num*vec.y  , "z" : num*vec.z };
	else
		return {"x" : num*vec.x  ,"y" : num*vec.y  };
}

function cos(a,b)
{
	return dot_prod(a,b)/(len(a)*(len(b)));
}

function len(a)
{
	return Math.sqrt(dot_prod(a,a));
}

function dot_prod(a,b)
{
	if(a.z !== undefined && !isNaN(a.z))
		return (a.x)*(b.x) + (a.y)*(b.y) + (a.z)*(b.z);
	else
		return (a.x)*(b.x) + (a.y)*(b.y);
}

function dist_2(a,b)
{
	return Math.sqrt((a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y));
}

function TriArea(a,b,c)
{
	const side1 = vec_sub(b,a);
	const side2 = vec_sub(c,b);
	const side3 = vec_sub(a,c);
	const perimeter = (side1 + side2 + side3)/2;
	return Math.sqrt(perimeter*((perimeter-side1)*(perimeter-side2)*(perimeter-side3)));
}

function isZero(num)
{
	return Math.abs(num) < 0.0000001;
}

function randomColor()
{
  const randomInt = (min, max) => {
    return Math.floor(Math.random() * (max - min + 1)) + min;
  };
    var h = randomInt(0, 360);
    var s = randomInt(42, 98);
    var l = randomInt(40, 90);
    return hsl2rgb(h,s,l);
}

function hsl2rgb (h, s, l) {

    var r, g, b, m, c, x

    if (!isFinite(h)) h = 0
    if (!isFinite(s)) s = 0
    if (!isFinite(l)) l = 0

    h /= 60
    if (h < 0) h = 6 - (-h % 6)
    h %= 6

    s = Math.max(0, Math.min(1, s / 100))
    l = Math.max(0, Math.min(1, l / 100))

    c = (1 - Math.abs((2 * l) - 1)) * s
    x = c * (1 - Math.abs((h % 2) - 1))

    if (h < 1) {
        r = c
        g = x
        b = 0
    } else if (h < 2) {
        r = x
        g = c
        b = 0
    } else if (h < 3) {
        r = 0
        g = c
        b = x
    } else if (h < 4) {
        r = 0
        g = x
        b = c
    } else if (h < 5) {
        r = x
        g = 0
        b = c
    } else {
        r = c
        g = 0
        b = x
    }

    m = l - c / 2
    r = Math.round((r + m) * 255)
    g = Math.round((g + m) * 255)
    b = Math.round((b + m) * 255)

    return { r: r, g: g, b: b };
}
