'use-strict';

window.onload = function(){
	init();
}

// monkey-patch dat.GUI

dat.GUI.prototype.removeFolder = function (fldl) {
	var name = fldl.name;
	var folder = this.__folders[name];
	if (!folder) {
		return;
	}
	folder.close();
	this.__ul.removeChild(folder.domElement.parentNode);
	delete this.__folders[name];
	this.onResize();
}

// global variables
var lastid = null;
var gui;
var colors_array =  generate_colors("#ffd700","#f01104",10);
var random_colors_array = generate_colors_array(40);
var bbox = new THREE.Box3();

var params = {
	example_name: '4',
	material: 'phongSmooth',
	mesh_color: '#0080ff',
	surface: true,
	curves_color: '#ffffff',
	curves: true,
	background_color: '#3a3a3a',
	wireframe : false,
	random_color : false,
	color_gradient : false,
	color_1 : '#6b85ff',
	color_2 : '#20207a',
	gradient_step : 3,
	grid_size : 30,
	light_config : '8 cornes',
	light_intensity : 0.45,
	scale_x : (1.00).toFixed(2),
	scale_y : (1.00).toFixed(2),
	scale_z : (1.00).toFixed(2),
	screenshot_resolution_str : "800x600",
	toIMG : Screenshot_IMG,
	toSVG : Screenshot_SVG
};

var fopen;
var loadManager;
var objLoader;

var examples_map = new Map();
var lights_config = new Map();

//threejs globals
var scene, camera , renderer, controls;
var center = new THREE.Vector3(0,0,0); // obj center
var size = new THREE.Vector3(10,10,10); // obj size

// three js init
function init() 
{
	if (!Detector.webgl)
		Detector.addGetWebGLMessage();

	// renderer
	renderer = new THREE.WebGLRenderer({antialias : true,preserveDrawingBuffer: true});
	renderer.setSize( window.innerWidth, window.innerHeight );
	renderer.setClearColor(currentParams.background_color);
	document.body.appendChild( renderer.domElement );
	// CAMERA
	camera = new THREE.PerspectiveCamera( 60,window.innerWidth/ window.innerHeight, 0.01, 10000 );
	camera.up = new THREE.Vector3(0, 0, 1);
	controls = new THREE.OrbitControls( camera, renderer.domElement );

	// world
	scene = new THREE.Scene();

	// ambient light
	var light = new THREE.AmbientLight( 0x666666);
	scene.add( light );

	fopen = document.createElement('input');
	fopen.type = 'file';
	fopen.accept = '.txt';
	fopen.multiple = '';
	fopen.style.visibility = 'hidden';

	window.addEventListener( 'resize', onWindowResize, false );
	params.screenshot_resolution_str = window.innerWidth+"x"+window.innerHeight;
	gui = new dat.GUI();
	gui.add(params, 'example_name', example_names).name("example").onChange(changeExampleGeometry).listen();
	gui.add(params, 'light_config', light_names).name("Lights config").onChange(changeLights).listen();
	gui.add(params, 'light_intensity',0,1).step(0.05).name("Lights Intensity").onChange(changeLightsIntensity);	
	gui.add(params, 'grid_size',8,60).step(1).name("Bezier Grid size").onChange(createDefaultGeometry);
	gui.add(params, 'material', materialNames).onChange(changeSurfaceMaterial);
	gui.addColor(params, 'mesh_color').name('surface color').onChange(changeSurfaceColor);
	gui.add(params, 'wireframe').name('wireframe').onChange(ShowHideWireframe);
	gui.add(params, 'surface').name("surface").onChange(ShowHideSurface);
	gui.addColor(params, 'curves_color').name('curves color').onChange(changeCurvesColor);
	gui.add(params, 'curves').name("curves").onChange(ShowHideCurves);
	gui.addColor(params, 'background_color').name('background color').onChange(changeBackgroundColor);
	gui.add(params, 'random_color').name("Random Color").onChange(changeSurfaceColor_Random);
	gui.add(params, 'color_gradient').name("Color gradient").onChange(changeSurfaceColor_Gradient);
	gui.addColor(params, 'color_1').name("Color Start").onChange(changeSurfaceColor_Gradient);
	gui.addColor(params, 'color_2').name("Color End").onChange(changeSurfaceColor_Gradient);
	gui.add(params, 'gradient_step',1,100).step(1).name("Gradient Steps").onChange(changeSurfaceColor_Gradient);
	gui.add(params, 'scale_x').name("Scale x").onChange(scaleScene);	
	gui.add(params, 'scale_y').name("Scale y").onChange(scaleScene);	
	gui.add(params, 'scale_z').name("Scale z").onChange(scaleScene);	
	gui.add(params, 'screenshot_resolution_str').name('Image Resolution');
	gui.add(params, 'toIMG').name('Take Screenshot');
	gui.add(params, 'toSVG').name('Scene to SVG');
	
	examples_map.set("4",examples[0]);
	examples_map.set("5",examples[1]);
	examples_map.set("25",examples[2]);
	examples_map.set("c4",examples_c[0]);
	examples_map.set("c5",examples_c[1]);
	examples_map.set("c25",examples_c[2]);
	
	lights_config.set('8 cornes',{'count':8,'apply_function':corners_light_apply});
	lights_config.set('Up-Down',{'count':2,'apply_function':up_down_light_apply});
	
	createMaterials();
	changeSurfaceColor();
	createDefaultGeometry();
	onWindowResize();
	animate();
}

function corners_light_apply(center,size,dist=10,type = "DirectionalLight")
{
	Array.from(scene.children.filter(x=>x.type.includes("Light") && x.type!=="AmbientLight")).forEach(l=>scene.remove(l));
	for(var i=0;i<8;i++)
	{
		var light;
		if(type == "DirectionalLight")
			light = new THREE.DirectionalLight( 0xffffff);
		else
			light = new THREE.SpotLight( 0xffffff);
		light.intensity = params.light_intensity;
		light.position.set( 0, 0, 0 );
		scene.add( light );
	}
	var l = Array.from(scene.children.filter(x=>x.type.includes("Light") && x.type!=="AmbientLight" ));
	var index = 0;
	for(var i=-1;i<=1;i=i+2)
	{
		for(var j=-1;j<=1;j=j+2)
		{
			for(var k=-1;k<=1;k=k+2)
			{
				var pos = new THREE.Vector3(center.x + i*0.5*size.x,center.y+ j*0.5*size.x,center.z+ k*0.5*size.x);
				var dir = (pos.clone()).sub(center);
				dir.normalize();
				dir.multiplyScalar(dist);
	
				l[index++].position.set(pos.x + dir.x,pos.y + dir.y ,pos.z + dir.z);				
			}
		}
	}
}

function up_down_light_apply(center,size,dist = 10)
{
	Array.from(scene.children.filter(x=>x.type.includes("Light") && x.type!=="AmbientLight" )).forEach(l=>scene.remove(l));
	for(var i=0;i<2;i++)
	{
		var light = new THREE.DirectionalLight( 0xffffff); 
		light.intensity = params.light_intensity;
		light.position.set( 0, 0, 0 );
		scene.add( light );
	}
	var l = Array.from(scene.children.filter(x=>x.type.includes("Light") && x.type!=="AmbientLight"));
	l[0].position.set(center.x,center.y,center.z + size.z*dist);
	l[1].position.set(center.x,center.y,center.z - size.z*dist);
}

// fit to size
function fitCameraToBoundingBox ( camera, boundingBox, offset, controls) 
{
    offset = offset || 1.25;
  
	boundingBox.getCenter(center);
	boundingBox.getSize(size);

    // get the max side of the bounding box (fits to width OR height as needed )
    const maxDim =  Math.max(size.y, size.z ) * 0.72;
    const fov = camera.fov * ( Math.PI / 180 );
    let cameraX_offset = Math.abs( maxDim / Math.tan( fov * 0.5 ) );

    cameraX_offset *= offset; // zoom out a little so that objects don't fill the screen

    camera.position.set(center.x-size.x*0.5-cameraX_offset,center.y,center.z+0.3*size.z);
    if ( controls ) {
      // set camera to rotate around center of loaded object
      controls.target.set(center.x,center.y,center.z);
	  controls.zoomSpeed =  Math.max(size.x,maxDim/0.72);
      controls.update();

    } 
	else 
        camera.lookAt( center );
	
	lights_config.get(params.light_config).apply_function(center,size,Math.max(size.x,maxDim/0.72)*10);
	
}

// Render
function animate() {
	render();
	lastid = requestAnimationFrame(animate);
	controls.update();
}

function onWindowResize() {
	camera.aspect = window.innerWidth / window.innerHeight;
	camera.updateProjectionMatrix();
	renderer.setSize( window.innerWidth, window.innerHeight );
	render();
}

function render() 
{
	renderer.render( scene, camera );
}

function ComputeSurface(input)
{
	var control_polygons_cubic = EdgesToBezier3(input.points,input.edge_info);
	var control_polygons_quartic = Bezier3To4(control_polygons_cubic);
	var triangles = TriangulationToArray(input.edge_info,input.triangulation,control_polygons_cubic,control_polygons_quartic);
	FindAlphaBeta(input.edge_info,triangles,control_polygons_quartic);
	var TS = FindTS(input.points,input.edge_info,triangles,input.triangulation,control_polygons_quartic);
	UpdateAlphaBeta(input.edge_info,triangles,control_polygons_quartic);
	Alg_1_Step_3_4(input.points,input.edge_info,triangles,control_polygons_cubic,control_polygons_quartic);
	SplittingVlachkova(triangles);	
	var bezier_surfaces_cp = triangles.map(t=>TrinagleToBezierControlPoints(t,control_polygons_quartic)).flat();
	return {'surface_control_polygons':bezier_surfaces_cp,
			'quartic_curves': control_polygons_quartic.map(cp=>BezierCurve(cp)) };
}

var example_names = [
	'4',
	'5',
	'25',
	'c4',
	'c5',
	'c25',
	'Load from TXT...',
	'Load CONVEX from TXT...',
];

var light_names = [
	'8 cornes',
	'Up-Down',
	'Load from TXT...'
];

var predefined_examples = [];

var materialNames = [
	'phongFlat',
	'phongSmooth',
	'lambert',
	'normal',
];

var materials = [];

var currentParams = {
	example_name: params.example_name,
	surface_meshes : [],
	curves_meshes: [],
	curves_mat: null,
	surface_color: new THREE.Color(parseInt(params.mesh_color.replace('#', '0x'))),
	curves_color: new THREE.Color(parseInt(params.curves_color.replace('#', '0x'))),
	background_color: new THREE.Color(parseInt(params.background_color.replace('#', '0x'))),
	material: params.material,
	
};


// Change events
function createDefaultGeometry( ) 
{
	params.scale_x = 1.00; params.scale_y = 1.00; params.scale_z = 1.00;
	//remove old scene
	currentParams.surface_meshes.forEach(function(m){
		m.geometry.dispose();
		m.material.dispose();
		scene.remove(m);
	});
	currentParams.curves_meshes.forEach(function(m){
		m.geometry.dispose();
		m.material.dispose();
		scene.remove(m);
	});
	var s = ComputeSurface(examples_map.get(params.example_name));
	bbox = new THREE.Box3();
	// surface 
	s.surface_control_polygons.forEach(function(surface_control_points){
		const mesh = BezierTrianglesMesh(surface_control_points,4,params.grid_size);
		var surface_geom = new THREE.Geometry();
		for(var index = 0;index<mesh.points.x.length;index++)
		{
			surface_geom.vertices.push(new THREE.Vector3(mesh.points.x[index], mesh.points.y[index], mesh.points.z[index]));
		}
		for(var index = 0;index<mesh.edges.i.length;index++)
		{
			surface_geom.faces.push(new THREE.Face3(mesh.edges.i[index], mesh.edges.j[index], mesh.edges.k[index]));
		}
		surface_geom.computeVertexNormals();
		surface_geom.computeFaceNormals();
		
		currentParams.surface_meshes.push(new THREE.Mesh(surface_geom));
		changeSurfaceMaterial();
		scene.add(currentParams.surface_meshes[currentParams.surface_meshes.length-1]);	
	});
	// curves
	s.quartic_curves.forEach(function(curve){
		var quartic_curve_geom = new THREE.Geometry();
		curve.forEach(function(p){
			quartic_curve_geom.vertices.push(new THREE.Vector3(p.x, p.y, p.z));
			bbox.expandByPoint(new THREE.Vector3(p.x, p.y, p.z)); 
		});

		currentParams.curves_meshes.push(new THREE.Line(quartic_curve_geom, currentParams.curves_mat ));
		scene.add(currentParams.curves_meshes[currentParams.curves_meshes.length-1]);
	});
//	console.log(bbox);
	fitCameraToBoundingBox(camera,bbox,1.1,controls);
	ShowHideSurface();
	ShowHideCurves();
	ShowHideWireframe();
}

function createMaterials() {
	var commonPhongParams = {
		color: currentParams.surface_color,
		shininess: 40,
		specular: 0x222222,
		side: THREE.DoubleSide,
	};
	materials['phongFlat'] = new THREE.MeshPhongMaterial(commonPhongParams);
	materials['phongFlat'].flatShading  = true;
	materials['phongSmooth'] = new THREE.MeshPhongMaterial(commonPhongParams);
	materials['phongSmooth'].flatShading  = false;
	materials['lambert'] = new THREE.MeshLambertMaterial(
	{
		color: currentParams.surface_color,
		side: THREE.DoubleSide,
	});
	materials['normal'] = new THREE.MeshNormalMaterial({side: THREE.DoubleSide});
	// create the wireframe material
	currentParams.curves_mat = new THREE.LineBasicMaterial({
		color: currentParams.curves_color,
		linewidth: 2,
		linecap: 'round', //ignored by WebGLRenderer
		linejoin:  'round' //ignored by WebGLRenderer
	});
}

function scaleScene()
{	
	if(params.scale_x <= 0 )
	{
		params.scale_x = 0.01; 
	}
	else if(params.scale_y <= 0)
	{
		params.scale_y = 0.01;
	}
	else if(params.scale_z <= 0)
	{
		params.scale_z = 0.01;
	}
	var x = params.scale_x; var y = params.scale_y; var z = params.scale_z;
	scene.children.forEach(o=>o.scale.set(x,y,z));	
	var min = bbox.min.clone(); var max = bbox.max.clone();
	min.x = min.x * x; min.y = min.y * y; min.z = min.z * z;
	max.x = max.x * x; max.y = max.y * y; max.z = max.z * z;
	var new_bbox = new THREE.Box3(min,max);
	fitCameraToBoundingBox(camera,new_bbox,1.1,controls);
}

function changeExampleGeometry() 
{
	//remove old scene
	currentParams.surface_meshes.forEach(function(m){
		m.geometry.dispose();
		m.material.dispose();
		scene.remove(m);
	});
	currentParams.curves_meshes.forEach(function(m){
		m.geometry.dispose();
		m.material.dispose();
		scene.remove(m);
	});
	currentParams.example_name = "";
	if (params.example_name == 'Load from TXT...' || params.example_name == 'Load CONVEX from TXT...') 
	{
		fopen.onchange = onFile;
		fopen.click();	
	}
	else 
	{
		createDefaultGeometry();
		currentParams.example_name  = params.example_name;
	}
}

function changeLights()
{
	if (params.light_config == 'Load from TXT...') 
	{
		fopen.onchange = onLights_File;
		fopen.click();
	}
	else 
	{
		lights_config.get(params.light_config).apply_function(center,size,Math.max(size.x,size.y,size.z)*10);
	}
}

function changeLightsIntensity()
{
	scene.children.filter(x=>x.type=="DirectionalLight").forEach(x=>x.intensity=params.light_intensity);
}

function onLights_File(event) 
{
	var input = event.target;
	var reader = new FileReader();
	reader.onload = function()
	{
		var text = reader.result;
		var filename = input.files[0].name.split(".").slice(0,-1).join(".");
		var lights = text.split("\n").map(function(l){
			var v = l.split(",");
			return 	{'pos': new THREE.Vector3(parseFloat(v[0]),parseFloat(v[1]),parseFloat(v[2])),
					'intensity' : parseFloat(v[3]),
					'color' : parseInt(v[4].replace("#","0x"))			
					};
		});
		console.log(lights);
		lights_config.set(filename,{'count':lights.length,'apply_function':function(){
			Array.from(scene.children.filter(x=>x.type=="DirectionalLight")).forEach(l=>scene.remove(l));
			for(var i=0;i<lights.length;i++)
			{
				var light = new THREE.DirectionalLight( lights[i].color , lights[i].intensity);
				light.position.set( lights[i].pos.x, lights[i].pos.y, lights[i].pos.z );
				scene.add( light );
			}
			
		}});
		light_names.splice(0,0,filename);
		params.light_config = filename;
		var c = document.createElement("option");
		c.value = filename;
		c.innerText = filename;
		gui.__controllers[1].domElement.querySelector("select").insertBefore(c,gui.__controllers[1].domElement.querySelector("select").firstElementChild);
		lights_config.get(filename).apply_function();
		
	};
	reader.readAsText(input.files[0]);
}

function onFile(event) 
{
	var input = event.target;
	var reader = new FileReader();
	reader.onload = function()
	{
		var text = reader.result;
		var filename = input.files[0].name.split(".").slice(0,-1).join(".");
		var input_obj = params.example_name == 'Load from TXT...' ? StringInputToArrays(text) : StringInputToArrays_C(text);
		examples_map.set(filename,input_obj);
		example_names.splice(0,0,filename);
		currentParams.example_name = filename;
		params.example_name = filename;
		var c = document.createElement("option");
		c.value = filename;
		c.innerText = filename;
		gui.__controllers[0].domElement.querySelector("select").insertBefore(c,gui.__controllers[0].domElement.querySelector("select").firstElementChild);
		createDefaultGeometry();
		
	};
	reader.readAsText(input.files[0]);
}

function changeSurfaceMaterial() {
	currentParams.surface_meshes.forEach(function(m){
		var back_up_color = null;
		try
		{
			back_up_color = m.material.color;
		}
		catch(err) { }
		m.material = null;
		m.material = materials[params.material].clone();
		if(materials[params.material].hasOwnProperty("color") && back_up_color!= null && (params.color_gradient == true || params.random_color == true))
		{
			m.material.color = back_up_color;
		}
			
	});
	currentParams.material = params.material;
	currentParams.surface_meshes.forEach(m=>m.material.needsUpdate = true);
	changeSurfaceColor_Gradient();
}

function changeSurfaceColor_Random()
{ 
	if(params.random_color == true)
	{
		params.color_gradient = false;
		
		var i = 0; var step = 1;
		currentParams.surface_meshes.forEach(function(s){
			if(i >= random_colors_array.length-1)
				step = -1;
			if(i == 0)
				step = 1;
			s.material.color = random_colors_array[i];
			i = i+step;
			s.material.needsUpdate = true;
		});
	}
	else
	{
		changeSurfaceColor();
	}
	for (var i in gui.__controllers) {
		gui.__controllers[i].updateDisplay();
	}
}

function changeSurfaceColor_Gradient()
{
	if(params.color_gradient == true)
	{
		params.random_color = false;
		colors_array = generate_colors(params.color_1,params.color_2,params.gradient_step);
		for(var i=0;i<currentParams.surface_meshes.length;i++)
		{ 
			currentParams.surface_meshes[i].material.color = new THREE.Color(parseInt(colors_array[i%colors_array.length].replace('#', '0x')));
		}
		currentParams.surface_meshes.forEach(s=>s.material.needsUpdate = true);
	}
	else
	{
		changeSurfaceColor();
	}
	for (var i in gui.__controllers) {
		gui.__controllers[i].updateDisplay();
	}
}

function changeSurfaceColor() 
{
	currentParams.mesh_color = new THREE.Color(parseInt(params.mesh_color.replace('#', '0x')));
	materials['phongFlat'].color = currentParams.mesh_color;
	materials['phongSmooth'].color = currentParams.mesh_color;
	materials['lambert'].color = currentParams.mesh_color;
	currentParams.surface_meshes.forEach(function(s){
		if(materials[params.material].hasOwnProperty("color") && params.color_gradient == false && params.random_color == false)
		{
			s.material.color = currentParams.mesh_color;
		}
		s.material.needsUpdate = true;
	});
}

function changeCurvesColor() {
	currentParams.curves_color = new THREE.Color(parseInt(params.curves_color.replace('#', '0x')));
	currentParams.curves_mat.color = currentParams.curves_color;
	currentParams.curves_mat.needsUpdate = true;
}

function changeBackgroundColor() {
	currentParams.background_color = new THREE.Color(parseInt(params.background_color.replace('#', '0x')));
	renderer.setClearColor(currentParams.background_color);
}

function ShowHideSurface() {
	currentParams.surface_meshes.forEach(m=>m.visible = params.surface);
}

function ShowHideCurves() {
	currentParams.curves_meshes.forEach(c=>c.visible = params.curves);
}

function ShowHideWireframe()
{
	for (var prop in materials) 
	{
		if (materials.hasOwnProperty(prop)) 
		{
			materials[prop].wireframe = params.wireframe;
		}
	}
	currentParams.surface_meshes.forEach(function(m){
		m.visible = params.surface;
		m.material.wireframe = params.wireframe;
	});
}

function generate_colors_array(size)
{
	var arr = [];
	for(var i = 0 ;i<size;i++)
	{
		var g = new THREE.Geometry();
		g.vertices.push(new THREE.Vector3(10000,100000,100000));
		g.vertices.push(new THREE.Vector3(1003000,10002000,10040));
		g.vertices.push(new THREE.Vector3(1003000,100200,103000));
		g.faces.push(new THREE.Face3(0, 1, 2));
		arr.push(new THREE.Mesh(g).material.color);
		g.dispose();		
	}
	return arr;
}

function Screenshot_IMG(format = "jpg")
{ 	
	cancelAnimationFrame(lastid);	
	var element = document.createElement('a');
	var resolution = params.screenshot_resolution_str.match(/(\d+)\s*x\s*(\d+)/i);
	if(!(resolution != null && resolution.length >=3))
	{
		resolution = ["800x600","800","600"];
	}
	renderer.setSize(  resolution[1], resolution[2] );
	camera.aspect = resolution[1]/resolution[2];
	camera.updateProjectionMatrix();
	render();
	var url = renderer.domElement.toDataURL("image/"+format);
	element.href =  url;
	element.setAttribute('download', "screenshot"+currentParams.example_name+"."+format);
	element.style.display = 'none';
	document.body.appendChild(element);

	element.click();

	document.body.removeChild(element);
	onWindowResize();
	requestAnimationFrame(animate)
}

function Screenshot_SVG(q = "high")
{ 	
	
	var XMLS = new XMLSerializer(); 
	cancelAnimationFrame(lastid);
	var renderer_svg =  new THREE.SVGRenderer();
	renderer_svg.setSize( window.innerWidth, window.innerHeight );
	renderer_svg.overdraw = 0.3;
	renderer_svg.setPrecision(12);
	renderer_svg.setQuality(q);
	renderer_svg.setClearColor(currentParams.background_color);
	renderer_svg.render( scene, camera );	
	var svgfile = XMLS.serializeToString(renderer_svg.domElement);
	var element = document.createElement('a');
	var data_blob = new Blob([svgfile], { type: 'data:image/svg+xml' }); 
	var url = URL.createObjectURL(data_blob);
	element.href =  url;
	element.setAttribute('download', "screenshot"+currentParams.example_name+".svg");
	element.style.display = 'none';
	document.body.appendChild(element);

	element.click();

	document.body.removeChild(element);
	requestAnimationFrame(animate)
}